/*
    Copyright (C) 2010 wpitchoune@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <string.h>

#include <sensors/sensors.h>
#include <sensors/error.h>

#include "psensor.h"
#include "nvidia.h"

void psensor_free(struct psensor* sensor)
{
	if (sensor) {
		free(sensor->name);
		free(sensor->chip_name);
		
		free(sensor);
	}
}

void psensor_list_free(struct psensor** sensors) 
{
	struct psensor** sensor_cur;

	if (sensors) {
		sensor_cur = sensors;
		
		while(*sensor_cur) {
			psensor_free(*sensor_cur);
			
			sensor_cur++;
		}
		
		free(sensors);
	}
}

/*
 * Returns minimal stored temperature
 */
int get_min_temp(struct psensor** sensors) 
{
	int m = 0;
	struct psensor** s = sensors;

	while(*s) {
		if ((*s)->enabled) {
			int i;
			int t;
			for(i=0;i<MAX_TEMPS;i++) {
				t = (*s)->temps[i];
				
				if (t && (t < m || m == 0)) 
					m = t;
			}
		}
		s++;
	}

	return m;
}

/*
 * Returns maximal stored temperature
 */
int get_max_temp(struct psensor** sensors) 
{
	int m = 0;
	struct psensor** s = sensors;

	while(*s) {
		if ((*s)->enabled) {
			int i;
			int t;
			for(i=0;i<MAX_TEMPS;i++) {
				t = (*s)->temps[i];
				
				if (t > m) 
					m = t;
			}
		}
		s++;
	}

	return m;
}

double get_value(const sensors_chip_name *name,
		 const sensors_subfeature *sub)
{
	double val;
	int err;

	err = sensors_get_value(name, sub->number, &val);
	if (err) {
		fprintf(stderr, "ERROR: Can't get value of subfeature %s: %s\n",
			sub->name, sensors_strerror(err));
		val = 0;
	}
	return val;
}


double get_sensor_temp_input(const sensors_chip_name* chip,
			     const sensors_feature* feature)
{
	const sensors_subfeature *sf;
	
	sf = sensors_get_subfeature(chip, feature,
				    SENSORS_SUBFEATURE_TEMP_INPUT);
	if (sf) 
		return get_value(chip, sf);
	else
		return 0;
}

double get_temp_input(struct psensor* sensor)
{
	if (sensor->type == SENSOR_TYPE_LMSENSOR) 
		return get_sensor_temp_input(sensor->iname,
					     sensor->feature);
	else
		return nvidia_get_sensor_temp(sensor);
}


void update_psensor(struct psensor* sensor) 
{
	int i;
	for(i = 1; i < MAX_TEMPS; i++) 
		sensor->temps[i-1] = sensor->temps[i];
	
	sensor->temps[MAX_TEMPS-1] = get_temp_input(sensor);
}



struct psensor* create_sensor(const sensors_chip_name* chip,
			      const sensors_feature* feature) 
{
	char name[200];
	
	if (sensors_snprintf_chip_name(name, 200, chip) < 0)
		return NULL;

	const sensors_subfeature *sf 
		= sensors_get_subfeature(chip, feature,
					 SENSORS_SUBFEATURE_TEMP_FAULT);
	if (sf && get_value(chip, sf)) 
		return NULL;

	
	char *label;				
	if (!(label = sensors_get_label(chip, feature))) 
		return NULL;

	if (!get_sensor_temp_input(chip,feature)) 
		return NULL;

	
	struct psensor* psensor 
		= (struct psensor*)malloc(sizeof(struct psensor));
	psensor->chip_name = strdup(name);
	psensor->name = strdup(label);
	psensor->iname = chip;
	psensor->feature = feature;
	psensor->enabled = 1;
	psensor->type = SENSOR_TYPE_LMSENSOR;

	int t;
	for(t=0;t<MAX_TEMPS;t++) 
		psensor->temps[t] = 0;

	return psensor;
}


struct psensor** get_all_sensors() 
{
	// number of uniq colors
	#define COLORS_COUNT 7

	int colors[COLORS_COUNT][3] = { 
		{1,0,0}, 
		{0,1,0}, 
		{0,0,1}, 
		{1,0,1},
		{0,1,1},
		{1,1,0},
		{0,0,0}
	};

	struct psensor** psensors = NULL;
	int count = 0;
	const sensors_chip_name *chip;
	int chip_nr = 0;

	while ((chip = sensors_get_detected_chips(NULL, &chip_nr))) {
		const sensors_feature *feature;
		int i = 0;
		while ((feature = sensors_get_features(chip, &i))) {
			struct psensor* psensor;

			if (feature->type ==  SENSORS_FEATURE_TEMP
			    && (psensor = create_sensor(chip, feature))) {
				psensor->color[0] = colors[count%COLORS_COUNT][0];
				psensor->color[1] = colors[count%COLORS_COUNT][1];
				psensor->color[2] = colors[count%COLORS_COUNT][2];

				struct psensor** tmp_psensors;
				tmp_psensors 
					= malloc((count+1+1)*sizeof(struct psensor*));
				
				memcpy(tmp_psensors, 
				       psensors, 
				       count*sizeof(struct psensor*));

				tmp_psensors[count] = psensor;
				tmp_psensors[count+1] = NULL;
				
				free(psensors);
				psensors = tmp_psensors;
				count++;
			}	
		}
	}

	int nvidia_gpus_count = nvidia_init();
	int i;
	for( i = 0 ; i < nvidia_gpus_count; i++) {
		struct psensor* sensor = nvidia_create_sensor(i);
		struct psensor** tmp_psensors;
		tmp_psensors 
			= malloc((count+1+1)*sizeof(struct psensor*));
		
		memcpy(tmp_psensors, 
		       psensors, 
		       count*sizeof(struct psensor*));
		
		tmp_psensors[count] = sensor;
		tmp_psensors[count+1] = NULL;
		
		free(psensors);
		psensors = tmp_psensors;
		count++;
	}

	return psensors;
}
